/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <alsa/asoundlib.h>
#include <poll.h>
#include "muspd.h"

struct pollfd *alsa_pollfds;
size_t alsa_pollfds_len;
size_t alsa_frag_siz = 0;
size_t alsa_frame_siz;
snd_pcm_t *alsa_pcm = NULL;

static int alsa_metaparams(const struct meta *s, snd_pcm_hw_params_t *params)
{
	if (snd_pcm_hw_params_set_access(alsa_pcm, params,
				SND_PCM_ACCESS_RW_INTERLEAVED)) {
		perrf("failed to set ALSA PCM access");
		return 1;
	}
	if (snd_pcm_hw_params_set_channels(alsa_pcm, params, s->chan)) {
		perrf("failed to set ALSA PCM channels");
		return 1;
	}
	if (snd_pcm_hw_params_set_rate(alsa_pcm, params, s->rate, 0)) {
		perrf("failed to set ALSA PCM rate");
		return 1;
	}
	snd_pcm_format_t fmt;
	if (s->sign == 0) {
		switch (s->depth) {
			case 1:
				fmt = SND_PCM_FORMAT_U8;
				break;
			case 2:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_U16_BE :
					SND_PCM_FORMAT_U16_LE;
				break;
			case 3:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_U24_BE :
					SND_PCM_FORMAT_U24_LE;
				break;
			case 4:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_U32_BE :
					SND_PCM_FORMAT_U32_LE;
				break;
			default:
				return 1;
		}
	} else {
		switch (s->depth) {
			case 1:
				fmt = SND_PCM_FORMAT_S8;
				break;
			case 2:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_S16_BE :
					SND_PCM_FORMAT_S16_LE;
				break;
			case 3:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_S24_BE :
					SND_PCM_FORMAT_S24_LE;
				break;
			case 4:
				fmt = s->end == BIG ?
					SND_PCM_FORMAT_S32_BE :
					SND_PCM_FORMAT_S32_LE;
				break;
			default:
				return 1;
		}
	}
	if (snd_pcm_hw_params_set_format(alsa_pcm, params, fmt)) {
		perrf("failed to set ALSA PCM format");
		return 1;
	}

	return 0;
}

static int alsa_set(const struct meta *s)
{
	int res = 1;
	snd_pcm_hw_params_t *params;
	if (s == NULL)
		return 1;
	if (snd_pcm_hw_params_malloc(&params))
		return 1;
	/* Note that, although the documentation says nothing,
	 * this function returns negative on error
	 * but may return zero or positive on success. */
	if (snd_pcm_hw_params_any(alsa_pcm, params) < 0)
		goto end;
	if (alsa_metaparams(s, params))
		goto end;

	int dir;
	snd_pcm_uframes_t period_siz;
	if (snd_pcm_hw_params_set_period_size_last(alsa_pcm, params,
				&period_siz, &dir))
		goto end;
	snd_pcm_uframes_t buffer_siz;
	if (snd_pcm_hw_params_set_buffer_size_last(alsa_pcm, params,
				&buffer_siz))
		goto end;
	if (snd_pcm_hw_params(alsa_pcm, params))
		goto end;
	if (snd_pcm_hw_params_get_period_size(params, &period_siz, &dir))
		goto end;
	const ssize_t frame_siz = snd_pcm_frames_to_bytes(alsa_pcm, 1);
	if (frame_siz < 0)
		goto end;
	alsa_frame_siz = frame_siz;
	alsa_frag_siz = period_siz * alsa_frame_siz;
	res = 0;

#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: ALSA_FRAG_SIZ: %zu", alsa_frag_siz);
#endif

end:
	snd_pcm_hw_params_free(params);
	if (res)
		perrf("failed to set ALSA audio settings");
	return res;
}

int alsa_init(const struct meta *s)
{
	/* alsa_init is called several times not just start of the program. */
	if (alsa_deinit())
		return 1;
	const char * const dev = muspd_audio_device ?
		muspd_audio_device : "default";
	/* It appears from the source code,
	 * that ALSA alsays sets O_CLOEXEC
	 * when opening its device files. */
	if (snd_pcm_open(&alsa_pcm, dev, SND_PCM_STREAM_PLAYBACK, 0)) {
		perrf("failed to open ALSA device %s", dev);
		return 1;
	}
	const int n = snd_pcm_poll_descriptors_count(alsa_pcm);
	if (n <= 0)
		return 1;
	alsa_pollfds_len = n;
	alsa_pollfds = malloc(alsa_pollfds_len * sizeof *alsa_pollfds);
	if (!alsa_pollfds) {
		alsa_pollfds_len = 0;
		return 1;
	}
	if (snd_pcm_poll_descriptors(alsa_pcm, alsa_pollfds,
				alsa_pollfds_len) != alsa_pollfds_len) {
		alsa_deinit();
		return 1;
	}
	if (alsa_set(s)) {
		alsa_deinit();
		return 1;
	}
	if (alsa_frag_siz == 0)
		return 1;

	return 0;
}

int alsa_deinit()
{
	if (alsa_pcm) {
		alsa_pollunreg();
		free(alsa_pollfds);
		alsa_pollfds_len = 0;
		if (snd_pcm_drain(alsa_pcm))
			return 1;
		if (snd_pcm_close(alsa_pcm))
			return 1;
		alsa_pcm = NULL;
	}

	return 0;
}

ssize_t alsa_write(const char *buf, size_t siz)
{
#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: WRITE: %zu", siz);
#endif
	return alsa_frame_siz * snd_pcm_writei(alsa_pcm, buf,
			siz / alsa_frame_siz);
}

ssize_t alsa_writeall(const char *buf, size_t siz)
{
	size_t n = 0;
	ssize_t m;
	while (n < siz) {
		m = alsa_write(buf + n, siz - n);
		if (m < 0)
			return -1;
		n += m;
	}

	return n;
}

ssize_t alsa_getfrags()
{
	const snd_pcm_sframes_t f = snd_pcm_avail(alsa_pcm);
	if (f < 0)
		return -1;
	return f * alsa_frame_siz / alsa_frag_siz;
}

int alsa_pollreg(int (*hand)(int))
{
	size_t i;
	int res = 0;
	for (i = 0; i < alsa_pollfds_len; i++) {
		if ((res = poll_reg(alsa_pollfds[i].fd,
					alsa_pollfds[i].events, hand)))
			break;
	}
	if (res) {
		size_t j;
		for (j = 0; j < i; j++)
			poll_unreg(alsa_pollfds[j].fd);
	}

	return res;
}

void alsa_pollunreg()
{
	size_t i;
	for (i = 0; i < alsa_pollfds_len; i++) {
		poll_unreg(alsa_pollfds[i].fd);
	}
}

int alsa_pollisreg()
{
	size_t i;
	int res = 0;
	for (i = 0; i < alsa_pollfds_len; i++) {
		if ((res = poll_isreg(alsa_pollfds[i].fd)))
			break;
	}

	return res;
}
