/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>

/* Send SIGKILL to child and wait for death of child. */
int child_killwait(pid_t pid)
{
	if (kill(pid, SIGKILL)) {
		/* Sending SIGKILL to a zombie returns an error,
		 * but that's not a problem for us. */
		if (errno != ESRCH)
			return 1;
	}
	do {
		int s;
		errno = 0;
		if (waitpid(pid, &s, 0) == -1) {
			if (errno != EINTR)
				return 1;
		} else {
			break;
		}
	} while (1);

	return 0;
}

/* Do as dup2 but also close oldfd after. */
static int dup2_close(int oldfd, int newfd)
{
	if (newfd != oldfd) {
		if (dup2(oldfd, newfd) == -1)
			return 1;
		if (close(oldfd))
			return 1;
	}

	return 0;
}

/* For use in child_exec.
 * Create argv and argc to pass to exec
 * from variadic argument list.
 * Prog is the name of the program
 * which is to be put in argv[0]. */
static const char **varg_argv(const char *prog, va_list vargs)
{
	char *a;
	size_t argc;
	size_t argv_size = 2; // Guess 2 to start.
	const char **argv;

	argv = malloc(argv_size * sizeof *argv);
	if (argv == NULL)
		return NULL;

	argv[0] = prog;
	argc = 1;
	do {
		a = va_arg(vargs, char *);
		if (argc == argv_size) {
			argv_size += argv_size / 2 + 1;
			void * const p = realloc(argv,
					argv_size * sizeof *argv);
			if (p == NULL) {
				free(argv);
				return NULL;
			}
			argv = p;
		}
		argv[argc++] = a;
	} while (a != NULL);

	return argv;
}

/* Execute prog with arguments given variadically.
 * The first argument is the first not the zeroth;
 * that is, prog is automatically made the zeroth argument.
 * Also, the stdout of the program
 * is piped to the file descriptor *outfd.
 * The return value is the pid of the spawned child
 * or -1 on error. */
pid_t child_exec(const char *prog, int *outfd, ...)
{
	pid_t res = -1;
	va_list ap;
	int fds[2];
	const char **argv;

	va_start(ap, outfd);
	argv = varg_argv(prog, ap);
	va_end(ap);
	if (argv == NULL)
		return -1;

	if (pipe(fds))
		goto end;
	res = fork();
	switch (res) {
		case 0:
			if (
					close(fds[0]) ||
					dup2_close(fds[1], 1) ||
					/* There should be no harm
					 * in casting const to nonconst
					 * here. See POSIX. */
					execvp(prog, (char **)argv)
			)
				exit(1);
			break;
		case -1:
			/* Already have an error,
			 * so no point in detecting error in close. */
			close(fds[0]);
			close(fds[1]);
			goto end;
			break;
		default:
			if (close(fds[1])) {
				child_killwait(res);
				res = -1;
				goto end;
			}
			*outfd = fds[0];
			break;
	}

end:
	free(argv);
	return res;
}
