/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <sys/types.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include "muspd.h"

#ifndef MUSPDEC
#define MUSPDEC "muspdec"
#endif

/* dec_fd being -1 should indicate
 * that the decoder is not running.
 * dec_fd being not -1 should indicate
 * that we think the decoder is running;
 * though it may have exited already
 * and, once that be determined,
 * it shall be set to -1. */
int dec_fd = -1;
static pid_t dec_pid = -1;
static const char * const dec_prog = MUSPDEC;

int dec_start(const char *path)
{
	if (dec_fd != -1)
		dec_stop();
	dec_pid = child_exec(dec_prog, &dec_fd, path, (char *)NULL);
	if (dec_pid == -1) {
		perrf_errno("failed to start decoder");
		return 1;
	}

	return 0;
}

int dec_stop()
{
	if (close(dec_fd))
		return 1;
	dec_fd = -1;
	if (dec_pid != -1 && child_killwait(dec_pid)) {
		perrf_errno("failed to reap or kill decoder");
		return 1;
	}
	dec_pid = -1;

	return 0;
}

/* The buffer in question
 * is meant to fit four unsigned longs
 * separated by newlines.
 * I approximate log_10(2) by 1/3
 * and add 2 to account for the rounding error
 * in integer division.
 * The +1 is for the terminating newline.
 * The other +1 is to fit a null char
 * at the end of the buffer. */
#define META_BUFSIZE (\
	4 * (CHAR_BIT / 3 + 2 + 1) \
	+ ((CHAR_BIT * sizeof(unsigned)) / 3 + 2 + 1) \
	+ ((CHAR_BIT * sizeof(unsigned long long)) / 3 + 2 + 1) \
	+ 1 \
)

int dec_meta(const char *path, struct meta *ret)
{
	char buf[META_BUFSIZE];
	int child_out;
	char *ln;

	const pid_t pid = child_exec(dec_prog, &child_out, "-M", path,
			(char *)NULL);
	if (pid == -1) {
		perrf_errno("failed to execute decoder to determine metadata");
		return 1;
	}
	/* Note that the buffer's size is chosen judiciously. */
	const ssize_t buf_len = readall(child_out, buf, sizeof buf - 1);
	if (buf_len < 0) {
		perrf_errno(
			"could not read output of decoder in metadata mode");
		return 1;
	}
	buf[buf_len] = '\0';
	/* The child is no longer needed. */
	if (close(child_out))
		return 1;
	if (child_killwait(pid)) {
		perrf_errno("failed to reap or kill decoder in metadata mode");
		return 1;
	}

	/* Signedness */
	ln = strtok(buf, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->sign = strtoul(ln, NULL, 10);

	/* Endianness */
	ln = strtok(NULL, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->end = strtoul(ln, NULL, 10);

	/* Bits per sample */
	ln = strtok(NULL, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->depth = strtoul(ln, NULL, 10) / 8;

	/* Sample rate */
	ln = strtok(NULL, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->rate = strtoul(ln, NULL, 10);

	/* Channels */
	ln = strtok(NULL, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->chan = strtoul(ln, NULL, 10);

	/* Total samples */
	ln = strtok(NULL, "\n");
	if (ln == NULL)
		goto fmterr;
	ret->samples = strtoull(ln, NULL, 10);

	return 0;

fmterr:
	perrf("metadata supplied by decoder is not correctly formatted");
	return 1;
}
