/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <stdio.h>

#include "muspd.h"

static int if_fd = -1;
static const char *if_path = NULL;

int if_init(const char *path)
{
	if_path = path;
	if_fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (if_fd == -1) {
		perrf_errno("could not create socket");
		return 1;
	}
	/* It's important to close on exec. */
	int flags;
	flags = fcntl(if_fd, F_GETFD);
	if (flags < 0)
		return 1;
	flags |= FD_CLOEXEC;
	if (fcntl(if_fd, F_SETFD, flags) == -1)
		return 1;
	struct sockaddr_un sa;
	sa.sun_family = AF_UNIX;
	strncpy(sa.sun_path, path, sizeof sa.sun_path - 1);
	if (bind(if_fd, (struct sockaddr *)&sa, sizeof sa)) {
		perrf_errno("could not bind to %s", path);
		return 1;
	}
	if (poll_reg(if_fd, POLLIN, if_serve))
		return 1;

	return 0;
}

int if_deinit()
{
	poll_unreg(if_fd);
	if (close(if_fd))
		goto err;
	if (unlink(if_path))
		goto err;

	return 0;

err:
	perrf("failed to close or unlink unix domain socket file");
	return 1;
}

/* Extract command name and possibly the argument
 * from null-terminated buffer. */
static char *getcmd(char *buf, char **arg)
{
	char * const p = strtok(buf, " \t");
	if (!p)
		return NULL;
	char * const q = strtok(NULL, "");
	*arg = q;

	return p;
}

static int writestr(int fd, struct sockaddr_un *sa, const char *str)
{
	const ssize_t str_len = strlen(str);

	if (sendto(fd, str, str_len, 0,
				(struct sockaddr *)sa, sizeof *sa)
			!= str_len) {
		perrf_errno("failed to send datagram");
		return 1;
	}
	return 0;
}

static int writestatus(int fd, struct sockaddr_un *sa)
{
	return writestr(fd, sa, pb_status());
}

static int writelient(int fd, struct sockaddr_un *sa, size_t i)
{
	return writestr(fd, sa, li[i]);
}

static int writenum(int fd, struct sockaddr_un *sa, unsigned long long num)
{
	/* Fits any unsigned long long. */
	char num_str[(CHAR_BIT * sizeof(unsigned long long)) / 3 + 2 + 1];
	snprintf(num_str, sizeof num_str, "%llu", num);
	if (writestr(fd, sa, num_str))
		return 1;

	return 0;
}

static int writelist(int fd, struct sockaddr_un *sa)
{
	size_t i;
	if (writenum(fd, sa, li_len))
		return 1;
	for (i = 0; i < li_len; i++) {
		if (writelient(fd, sa, i))
			return 1;
	}

	return 0;
}

/* Receive datagram and act accordingly.
 * Note that we return 0 on some errors
 * as they should not cause the server to die. */
int if_serve(int fd)
{
	char buf[4096 + 1];
	struct sockaddr_un sa;
	socklen_t sa_len = sizeof sa;
	char *cmd, *arg;

	const ssize_t n = recvfrom(fd, buf, sizeof buf - 1, MSG_DONTWAIT,
			(struct sockaddr *)&sa, &sa_len);
	if (n <= 0) {
		perrf_errno("error receiving datagram");
		perrf("ignoring...");
		return 0;
	}
	buf[n] = '\0';
	cmd = getcmd(buf, &arg);
	if (cmd == NULL) {
		perrf("datagram contains no command");
		perrf("ignoring...");
		return 0;
	}
	if (strcmp(cmd, "play") == 0) {
		if (pb_play())
			goto err;

	} else if (strcmp(cmd, "pause") == 0) {
		if (pb_pause())
			goto err;

	} else if (strcmp(cmd, "stop") == 0) {
		if (pb_stop())
			goto err;

	} else if (strcmp(cmd, "list") == 0) {
		if (writelist(fd, &sa))
			goto cmderr;

	} else if (strcmp(cmd, "status") == 0) {
		if (writestatus(fd, &sa))
			goto cmderr;

	} else if (strcmp(cmd, "time") == 0) {
		if (writenum(fd, &sa, pb_time()))
			goto cmderr;

	} else if (strcmp(cmd, "duration") == 0) {
		if (writenum(fd, &sa, pb_duration()))
			goto cmderr;

	} else if (strcmp(cmd, "track") == 0) {
		if (writenum(fd, &sa, pb_track))
			goto cmderr;

	} else if (strcmp(cmd, "die") == 0) {
		exit(0);
	} else if (arg) {
		/* Commands taking an argument */
		char *p;
		unsigned long index;
		if (strcmp(cmd, "start") == 0) {
			index = strtoul(arg, &p , 10);
			if (*p || index > UINT_MAX)
				goto cmderr;
			if (index >= li_len)
				goto cmderr;
			if (pb_start(index, 1))
				goto err;
		} else if (strcmp(cmd, "add") == 0) {
			if (li_add(arg))
				goto cmderr;
		}
	} else {
		perrf("command not recognised or missing necessary argument");
		goto cmderr;
	}

	return 0;

/* Non fatal error */
cmderr:
	perrf("could not obey command");
	perrf("ignoring...");
	return 0;

/* Fatal error */
err:
	perrf("error executing command");
	return 1;
}
