/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include "muspd.h"

char *muspd_audio_device;

static void deinit()
{
	int res = 0;
	if (dec_stop())
		res = 1;
	if (if_deinit())
		res = 1;
	if (res)
		_exit(res);
}

static void sig_hand(int sig)
{
	deinit();

	struct sigaction sa = { 0 };
	sa.sa_handler = SIG_DFL;
	sigemptyset(&sa.sa_mask);
	sigaction(sig, &sa, NULL);
	raise(sig);
}

static int sig_init()
{
	struct sigaction act = { 0 };
	act.sa_handler = sig_hand;
	if (sigfillset(&act.sa_mask))
		return 1;
	if (sigaction(SIGINT, &act, NULL))
		return 1;
	if (sigaction(SIGTERM, &act, NULL))
		return 1;
	if (sigaction(SIGQUIT, &act, NULL))
		return 1;

	return 0;
}

static int daemonise()
{
	switch (fork()) {
		case 0:
			if (setsid() == -1)
				return 1;
			switch (fork()) {
				case 0:
					return 0;
				case -1:
					return 1;
				default:
					exit(0);
			}
		case -1:
			return 1;
		default:
			exit(0);
	}
}

static int env_init()
{
	char *str = getenv("MUSP_AUDIO_DEVICE");
	if (str) {
		muspd_audio_device = strdup(str);
		if (!muspd_audio_device)
			return 1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	perrf_head = argv[0];

	const char *unix;
	switch (argc) {
		case 1:
			unix = "muspd.unix";
			break;
		case 2:
			unix = argv[1];
			break;
		default:
			perrf("too many arguments");
			return 1;

	}

	if (env_init())
		return 1;
	if (poll_init())
		return 1;
	if (if_init(unix)) {
		perrf("failed to initialize interface");
		return 1;
	}
	if (li_init(0))
		return 1;
	if (sig_init())
		return 1;
	if (daemonise()) {
		perrf("failed to become a daemon");
		return 1;
	}
	if (atexit(deinit))
		return 1;

	return poll_loop();
}
