/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef MUSPD_H
#define MUSPD_H

#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <poll.h>
#include <perrf.h>

/* Audio metadata.
 * IF MODIFYING: also modify `metacmp' function. */
enum { LITTLE = 0, BIG };
struct meta {
	/* The first fould types are whaterver I like. */
	unsigned char		sign;
	unsigned char		end;
	unsigned char		depth;
	unsigned long long	samples;
	/* These last types are what OSS wants, so shouldn't be changed. */
	int			rate;
	int 			chan;
};

/* Main */
extern char *muspd_audio_device;

/* Polling */
int poll_init();
int poll_reg(int fd, int events, int (*hand)(int));
void poll_unreg(int fd);
int poll_isreg(int fd);
int poll_loop();

/* Decoder */
extern int dec_fd;
int dec_meta(const char *path, struct meta *ret);
int dec_start(const char *path);
int dec_stop();

#define MUSPD_AUDIO_OSS		0x0
#define MUSPD_AUDIO_ALSA	0x1

#ifndef MUSPD_AUDIO
#	error Choose audio backend by defining the MUSPD_AUDIO macro
#endif

#if MUSPD_AUDIO == MUSPD_AUDIO_OSS
#	define audio_frag_siz	oss_frag_siz
#	define audio_init	oss_init
#	define audio_deinit	oss_deinit
#	define audio_write	oss_write
#	define audio_writeall	oss_writeall
#	define audio_getfrags	oss_getfrags
#	define audio_pollreg	oss_pollreg
#	define audio_pollunreg	oss_pollunreg
#	define audio_pollisreg	oss_pollisreg
#elif MUSPD_AUDIO == MUSPD_AUDIO_ALSA
#	define audio_frag_siz	alsa_frag_siz
#	define audio_init	alsa_init
#	define audio_deinit	alsa_deinit
#	define audio_write	alsa_write
#	define audio_writeall	alsa_writeall
#	define audio_getfrags	alsa_getfrags
#	define audio_pollreg	alsa_pollreg
#	define audio_pollunreg	alsa_pollunreg
#	define audio_pollisreg	alsa_pollisreg
#else
#	error Unknown audio backend
#endif

#if MUSPD_AUDIO == MUSPD_AUDIO_OSS
/* OSS */
extern int oss_fd;
extern size_t oss_frag_siz;
int oss_init(const struct meta *s);
int oss_deinit();
ssize_t oss_write(const char *buf, size_t size);
ssize_t oss_writeall(const char *buf, size_t size);
ssize_t oss_getfrags();
int oss_pollreg(int (*hand)(int));
void oss_pollunreg();
int oss_pollisreg();
#endif

#if MUSPD_AUDIO == MUSPD_AUDIO_ALSA
/* ALSA */
extern size_t alsa_frag_siz;
int alsa_init(const struct meta *s);
int alsa_deinit();
ssize_t alsa_write(const char *buf, size_t size);
ssize_t alsa_writeall(const char *buf, size_t size);
ssize_t alsa_getfrags();
int alsa_pollreg(int (*hand)(int));
void alsa_pollunreg();
int alsa_pollisreg();
#endif

/* Interface */
int if_init(const char *path);
int if_deinit();
int if_serve(int fd);

/* Track List */
extern char **li;
extern size_t li_len;
int li_init();
int li_add(const char *track);
int li_clear();
int li_next();

/* Playback */
extern size_t pb_track;
int pb_osshand(int fd);
int pb_dechand(int fd);
int pb_start(size_t i, int trunc);
int pb_play();
int pb_pause();
int pb_next();
int pb_prev();
int pb_stop();
const char *pb_status();
unsigned long long pb_time();
unsigned long long pb_duration();

/* Child Control */
int child_killwait(pid_t pid);
int child_exec(const char *prog, int *outfd, ...);

/* Misc */
inline ssize_t readall(int fd, char *buf, size_t siz)
{
	size_t n = 0;
	ssize_t m;
	while (n < siz) {
		m = read(fd, buf + n, siz - n);
		if (m < 0)
			return -1;
		if (m == 0) {
			break;
		}
		n += m;
	}

	return n;
}
inline ssize_t writeall(int fd, const char *buf, size_t siz)
{
	size_t n = 0;
	ssize_t m;
	while (n < siz) {
		m = write(fd, buf + n, siz - n);
		if (m < 0)
			return -1;
		n += m;
	}

	return n;
}
inline int metacmp(const struct meta *m1, const struct meta *m2)
{
	if (m1->sign != m2->sign)
		return 1;
	if (m1->end != m2->end)
		return 1;
	if (m1->depth != m2->depth)
		return 1;
	if (m1->rate != m2->rate)
		return 1;
	if (m1->chan != m2->chan)
		return 1;
	/* We do not compare `samples'. */

	return 0;
}

#endif
