/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include "muspd.h"

int oss_fd = -1;
size_t oss_frag_siz = 0;

#if 0
/* These parameters determine
 * the internal buffer
 * of the sound card driver.
 * They don't have to be set,
 * but setting them can adjust the latencies.
 * OSS_FRAG is log_2 the size of a ``fragment''
 * OSS_FRAGNUM is the size of the buffer.
 * It's ideal not to have to set these manually.
 * I tried to choose some sensible values
 * for the most common type of audio I play (CDs). */
#define OSS_FRAG 0xe
#define OSS_FRAGNUM 0x5
/* I decided to let OSS set these automatically;
 * but the comments above and further down may be useful. */
#endif

static int oss_set(const struct meta *s)
{
	int res = 1;
	if (s != NULL) {
		if (ioctl(oss_fd, SNDCTL_DSP_CHANNELS, &s->chan))
			goto end;
		if (ioctl(oss_fd, SNDCTL_DSP_SPEED, &s->rate))
			goto end;
		int fmt;
		if (s->sign == 0) {
			if (s->depth == 1) {
				fmt = AFMT_U8;
			} else {
				/* Apparently,
				 * unsigned is, de facto, not a thing...
				 * Except for 8 bit,
				 * where signed is deprecated on OSS... */
				goto end;
			}
		} else {
			switch (s->depth) {
				case 1:
					fmt = AFMT_S8;
					break;
				case 2:
					fmt = s->end == BIG ?
						AFMT_S16_BE : AFMT_S16_LE;
					break;
#ifdef AFMT_S24_BE
#ifdef AFMT_S25_LE
				case 3:
					fmt = s->end == BIG ?
						AFMT_S24_BE : AFMT_S24_LE;
					break;
#endif
#endif
#ifdef AFMT_S32_BE
#ifdef AFMT_S32_LE
				case 4:
					fmt = s->end == BIG ?
						AFMT_S32_BE : AFMT_S32_LE;
					break;
#endif
#endif
				default:
					goto end;
			}
		}
		if (ioctl(oss_fd, SNDCTL_DSP_SETFMT, &fmt))
			goto end;
	}
	if (ioctl(oss_fd, SNDCTL_DSP_GETBLKSIZE, &oss_frag_siz))
		return 1;
	res = 0;

end:
	if (res)
		perrf_errno("failed to set OSS audio settings");
	return res;
}

int oss_init(const struct meta *s)
{
	/* oss_init is called several times not just start of the program. */
	if (oss_deinit())
		return 1;
	const char * const dev = muspd_audio_device ?
		muspd_audio_device : "/dev/dsp";
	/* It's important to close on exec. */
	oss_fd = open(dev, O_WRONLY | O_CLOEXEC);
	if (oss_fd == -1) {
		perrf_errno("failed to open OSS device %s", dev);
		return 1;
	}
#if 0
	/* The higher 16 bits
	 * indicate the "number of fragments"
	 * and the lower the "size of a fragment"
	 * (it's the log_2 of the size).
	 * My understanding of this is that
	 * there is a buffer in the driver
	 * with size being the number of fragments
	 * times the size of a frament.
	 * When data gets sent to the sound card,
	 * it is sent as a whole fragment;
	 * therefore, a small fragment size
	 * implies more CPU usage.
	 * Reducing the amount of fragments
	 * may cause the buffer to empty
	 * but should not impact CPU usage much. */
	int frag = (OSS_FRAGNUM << 16) | OSS_FRAG;
	if (ioctl(oss_fd, SNDCTL_DSP_SETFRAGMENT, &frag)) {
		perrf_errno("failed to set OSS fragment settings");
		return 1;
	}
#endif
	if (oss_set(s)) {
		oss_deinit();
		return 1;
	}
	if (oss_frag_siz == 0)
		return 1;

	return 0;
}

int oss_deinit()
{
	if (oss_fd != -1) {
		oss_pollunreg();
		if (close(oss_fd))
			return 1;
		oss_fd = -1;
	}

	return 0;
}

ssize_t oss_write(const char *buf, size_t size)
{
#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: WRITE: %zu", size);
#endif
	return write(oss_fd, buf, size);
}

ssize_t oss_writeall(const char *buf, size_t size)
{
#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: WRITE: %zu", size);
#endif
	return writeall(oss_fd, buf, size);
}

/* Return maximum number of full fragments
 * that can be written without blocking.
 * In case of error, return -1. */
ssize_t oss_getfrags()
{
	audio_buf_info bi;
	if (ioctl(oss_fd, SNDCTL_DSP_GETOSPACE, &bi))
		return -1;

	return bi.fragments;
}

int oss_pollreg(int (*hand)(int))
{
	return poll_reg(oss_fd, POLLOUT, hand);
}

void oss_pollunreg()
{
	poll_unreg(oss_fd);
}

int oss_pollisreg()
{
	return poll_isreg(oss_fd);
}
