/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

/* Here where is playback is orchestrated.  This is by far the most delicate
 * part of muspd.  Be careful before making any changes; I thought quite hard to
 * make this section of the program very efficient.  One immediately notices
 * that things could be much simpler, but, after extensive experimentation, I
 * believe this to be the ideal approach.  See further comments for
 * explanations. */

#include <stdlib.h>
#include <string.h>
#include "muspd.h"

size_t pb_track = 0;
/* playing iff `oss_fd or dec_fd registered'. */
static int playing = 0;
/* Number of bytes written to OSS */
static unsigned long long played = 0;
static struct meta met;

/* We use a pair of buffers, one for reading from dec, another for writing to
 * OSS.  Once wbuf is consumed, it is refilled with the contents of rbuf; should
 * it be possible, they are simply swapped for better performance.  There are
 * simpler approaches; this one has a few advantages. One of them is that we can
 * make the reads from dec larger than the writes to OSS; there is a speed
 * disparity between the two, so this is convenient. Another is that it helps
 * ensure writes to OSS are done in full fragments even at the end of a track,
 * provided the next has the same settings. */
static char *buf1 = NULL;
static char *buf2 = NULL;
static char **rbuf = &buf1;
static char **wbuf = &buf2;
static size_t bufs_siz = 0;
static size_t rbuf_len = 0;
static size_t wbuf_i = 0;
static size_t wbuf_len = 0;

/* This function ensures that
 * the size of the buffers is a multiple of an OSS fragment.
 * It also initialises the buffers on the first call. */
static int resizebufs()
{
	if (bufs_siz != audio_frag_siz) {
		bufs_siz = audio_frag_siz;
		void *p = realloc(buf1, bufs_siz);
		if (p == NULL)
			return 1;
		buf1 = p;
		p = realloc(buf2, bufs_siz);
		if (p == NULL)
			return 1;
		buf2 = p;
	}

	return 0;
}

static void swapbufs()
{
#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: SWAPBUFS");
#endif
	void * const p = rbuf;
	rbuf = wbuf;
	wbuf = p;
	wbuf_len = rbuf_len;
	wbuf_i = 0;
	rbuf_len = 0;
}

static void clearbufs()
{
	wbuf_len = 0;
	wbuf_i = 0;
	rbuf_len = 0;
}

/* Write both buffers to OSS. */
static int flushbufs()
{
#ifdef MUSPD_AUDIO_DEBUG
	perrf("DEBUG: FLUSHBUFS");
#endif
	if (audio_write(*wbuf + wbuf_i, wbuf_len - wbuf_i) < 0)
		goto err;
	swapbufs();
	if (audio_write(*wbuf + wbuf_i, wbuf_len - wbuf_i) < 0)
		goto err;
	wbuf_len = 0;

	return 0;

err:
	perrf_errno("error writing to sound card");
	return 1;
}

/* Determine if there is enough data across both buffers to constitute an OSS
 * fragment.  Note that we must be careful to avoid underflow as I like to use
 * unsigned integers. */
#define ENOUGH_FOR_FRAG() (rbuf_len + wbuf_len >= audio_frag_siz + wbuf_i)

/* Refill wbuf in order to have at least enough for a frament.  If wbuf be
 * empty, a fast swap of buffers is done, else, we need to copy.  The swap
 * should happen nearly always as the decoder is typically faster than OSS;
 * however, especially immediately after (re)starting playback, the transfers
 * from dec can be a lot slower as the kernel (linux) seems to use a smaller
 * pipe buffer initially. */
static void refillwbuf()
{
	/* If there not be sufficient data on the read buffer,
	 * we stop polling OSS. */
	if (!ENOUGH_FOR_FRAG()) {
		audio_pollunreg();
		return;
	} else if (wbuf_len == wbuf_i) {
		swapbufs();
	} else {
		const size_t siz = bufs_siz - wbuf_len > rbuf_len ?
				rbuf_len :
				bufs_siz - wbuf_len;
		memcpy(*wbuf + wbuf_len, *rbuf, siz);
		wbuf_len += siz;
		memmove(*rbuf, *rbuf + siz, rbuf_len - siz);
		rbuf_len -= siz;
	}
}

/* Poll handler for OSS */
int pb_audiohand(int fd)
{
	ssize_t frags = audio_getfrags();
	if (frags < 0)
		return 1;
	/* Unfortunately, at least using ALSA's OSS emulation on Linux, polling
	 * often returns with less than a full fragment free on the OSS
	 * buffer. */
	if (frags == 0) {
		/* It seems OSS does not block us for too long if we send a
		 * full framgment anyway. */
		frags = 1;
	}
	size_t out = frags * audio_frag_siz;
	/* It's recommendable to write full fragments. */
	if (out > wbuf_len - wbuf_i)
		out = ((wbuf_len - wbuf_i) / audio_frag_siz) * audio_frag_siz;
	if (out > 0) {
		const ssize_t n = audio_write(*wbuf + wbuf_i, out);
		if (n < 0) {
			perrf_errno("error writing to OSS");
			return 1;
		}
		wbuf_i += n;
		played += n;
	} else {
		refillwbuf();
	}
	/* pb_dechand stops the polling on dec_fd once rbuf be filled.
	 * If, now, there be space in rbuf, restart polling. */
	if (rbuf_len < bufs_siz && !poll_isreg(dec_fd)) {
		if (poll_reg(dec_fd, POLLIN, pb_dechand))
			return 1;
	}

	return 0;
}

/* Poll handler for decoder */
int pb_dechand(int fd)
{
	const size_t in = bufs_siz - rbuf_len;
	if (in > 0) {
		const ssize_t n = read(dec_fd, *rbuf + rbuf_len, in);
		if (n < 0) {
			perrf_errno("error reading data from decoder");
			return 1;
		}
		if (n == 0)
			return pb_next();
		rbuf_len += n;
	}
	/* We must unconditionally make sure oss_fd be registered.
	 * Else, we may end up with neither oss_fd nor dec_fd registered. */
	if (!audio_pollisreg()) {
		if (audio_pollreg(pb_audiohand))
			return 1;
	}
	/* If we filled the buffer, stop polling on dec_fd. */
	if (rbuf_len == bufs_siz)
		poll_unreg(dec_fd);

	return 0;
}

/* Set metadata to that of track. */
static int newmeta(size_t track)
{
	struct meta new_met;
	if (dec_meta(li[track], &new_met))
		return 1;
	/* We want to avoid doing this if not necessary
	 * as it involves closing and reopening /dev/dsp,
	 * which causes the audio playback to hiccup.
	 * Note that metacmp does not compare `samples'. */
	const int diff = metacmp(&new_met, &met);
	met = new_met;
	if (diff) {
		met = new_met;
		if (playing) {
			/* We must flush as the settings have changed. */
			if (flushbufs())
				return 1;
			/* Note that oss_pollunreg is called by oss_init. */
			if (audio_init(&met))
				return 1;
			if (audio_pollreg(pb_audiohand))
				return 1;
		} else {
			/* Since not playing,
			 * best we can do is throw away the buffers. */
			clearbufs();
			/* We init and deinit OSS in order
			 * to set oss_frag_siz. */
			if (audio_init(&met))
				return 1;
			if (audio_deinit(&met))
				return 1;
		}
		/* Note that this also initialises the buffers
		 * on the first call. */
		if (resizebufs())
			return 1;
	}

	return 0;
}

/* Start playing track i.
 * Argument trunc indicates whether the buffers should be truncated for quicker
 * transition to new track; trunc should be 0 except when running this function
 * as per user's request. */
int pb_start(size_t i, int trunc)
{
	if (i >= li_len)
		goto err;
	pb_track = i;
	if (trunc)
		clearbufs();
	if (newmeta(pb_track))
		goto err;
	if (!playing) {
		if (audio_init(&met))
			goto err;
		if (audio_pollreg(pb_audiohand))
			goto err;
	}
	if (dec_fd != -1) {
		poll_unreg(dec_fd);
		if (dec_stop())
			goto err;
	}
	if (dec_start(li[pb_track]))
		goto err;
	if (poll_reg(dec_fd, POLLIN, pb_dechand))
		goto err;
	playing = 1;
	played = 0;

	return 0;

err:
	perrf("could not start track %llu", i);
	return 1;
}

int pb_play()
{
	if (dec_fd == -1)
		return 1;
	if (!playing) {
		if (audio_init(&met))
			goto err;
		if (audio_pollreg(pb_audiohand))
			goto err;
		playing = 1;
	}

	return 0;

err:
	perrf("failed to resume playback");
	return 1;
}

int pb_pause()
{
	if (dec_fd == -1)
		return 1;
	if (playing) {
		/* Deinitialising OSS allows other programs
		 * to use the sound card while we be paused.
		 * Note that oss_pollunreg is called by oss_deinit. */
		if (audio_deinit())
			return 1;
		playing = 0;
	}

	return 0;
}

int pb_next()
{
	if (pb_track == li_len - 1) {
		/* Increasing the track past the end of the track list has the
		 * effect that if the user append more tracks and starts
		 * playback on the `current' track, he will have the expected
		 * result. */
		pb_track++;
		return pb_stop();
	}

	return pb_start(pb_track + 1, 0);
}

int pb_stop()
{
	if (playing) {
		if (audio_deinit())
			return 1;
		playing = 0;
	}
	if (dec_fd != -1) {
		poll_unreg(dec_fd);
		if (dec_stop())
			return 1;
	}

	return 0;
}

unsigned long long pb_time()
{
	const unsigned long long sample_size = met.chan * met.depth * met.rate;
	if (sample_size == 0)
		return 0;
	return played / sample_size;
}

unsigned long long pb_duration()
{
	if (met.rate == 0)
		return 0;
	return met.samples / met.rate;
}

const char *pb_status()
{
	if (dec_fd == -1)
		return "stopped";
	else if (playing)
		return "playing";
	else
		return "paused";
}
