/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <poll.h>
#include <errno.h>
#include <stdlib.h>
#include "muspd.h"

#define INIT_SIZE 3

static size_t size = 0;
static size_t len = 0;
static struct pollfd *fds;
static int (**hds)(int);

int poll_init()
{
	size = INIT_SIZE;
	fds = malloc(size * sizeof *fds);
	if (fds == NULL)
		goto err;
	hds = malloc(size * sizeof *hds);
	if (hds == NULL) {
		free(fds);
		goto err;
	}
	return 0;

err:
	perrf_errno("failed to allocate memory");
	return 1;
}

/* Grow arrays by at least one entry.
 * Note that, if failure happen on the second reallocation,
 * the first array will remain larger than the second
 * and `size' will reflect the size prior to calling this function.
 * In any event, `size' will reflect the minimum of the two sizes.
 * So, in case of failure,
 * the caller may not assume the arrays grew
 * but, otherwise, the arrays may continue to be used. */
static int grow()
{
	const size_t new_size = size + size / 2 + 1;
	void *p = realloc(fds, new_size * sizeof *fds);
	if (p == NULL)
		goto err;
	fds = p;
	p = realloc(hds, new_size * sizeof *hds);
	if (p == NULL)
		goto err;
	hds = p;

	size = new_size;

	return 0;

err:
	perrf_errno("failed to allocate memory");
	return 1;
}

/* Trim arrays to size `len'.
 * As opposed to `grow', here,
 * we set size prior to the second reallocation.
 * Thus, the effect is the same as for `grow':
 * if the second reallocation fail,
 * `size' will correctly reflect the minimum of the two sizes
 * and the arrays may continue to be used without problem. */
static int trim()
{
	void *p = realloc(fds, len * sizeof *fds);
	if (p == NULL)
		return 1;
	fds = p;
	size = len;
	p = realloc(hds, len * sizeof *hds);
	if (p == NULL)
		return 1;
	hds = p;

	return 0;
}

/* It's not a catastrophic error if a call be unsuccessful.
 * In that event, the caller can try to call again later. */
int poll_reg(int fd, int events, int (*hand)(int))
{
	if (len == size) {
		if (grow()) {
			perrf("failed to add filehandle to be polled");
			return 1;
		}
	}
	fds[len].fd = fd;
	fds[len].events = events;
	/* It is CRUCIAL to set revents to zero here.
	 * See poll_loop to see why. */
	fds[len].revents = 0;
	hds[len] = hand;
	len++;

	return 0;
}

void poll_unreg(int fd)
{
	size_t i, j;

	for (i = 0; i < len; i++) {
		if (fds[i].fd == fd)
			break;
	}
	if (i == len)
		return;
	for (j = i; j + 1 < len; j++) {
		/* It's important to copy the whole struct pollfd
		 * struct; in particular, revents MUST be copied.
		 * See poll_loop to see why. */
		fds[j] = fds[j + 1];
		hds[j] = hds[j + 1];
	}
	len--;

	if (len < size / 2) {
		/* Ignore failure
		 * since trim takes care to leave arrays and `size'
		 * in a usable state even in case of failure. */
		trim();
	}
}

int poll_isreg(int fd)
{
	size_t i;
	for (i = 0; i < len; i++) {
		if (fds[i].fd == fd)
			return 1;
	}

	return 0;
}

int poll_loop()
{
	int res;
	size_t i;
	while (1) {
		res = poll(fds, len, -1);
#ifdef MUSPD_AUDIO_DEBUG
		perrf("DEBUG: POLL");
#endif
		if (res < 0) {
			if (errno != EINTR) {
				perrf_errno("error polling");
				return 1;
			} else
				continue;
		}
		if (res == 0)
			continue;
		/* This loop is neat but a bit subtle.
		 * The reason is due to the possibility
		 * of the handlers (un)registering descriptors.
		 * After we run a given handler,
		 * we must restart the loop;
		 * we use revents to determine
		 * if the handler must be run;
		 * therefore, we set revents to zero
		 * on the descriptors that are about to be handled;
		 * also, recall that we set revents to zero
		 * on any newly registered descriptors;
		 * this prevents the handlers from being
		 * invoked on them. */
		for (i = 0; i < len; i++) {
			if (fds[i].revents) {
				/* It's important to set this to zero
				 * before the call to the handler
				 * as the meaning of `i' becomes
				 * lost right after the call. */
				fds[i].revents = 0;
				if (hds[i](fds[i].fd)) {
					perrf("exiting...");
					return 1;
				}
				i = 0;
			}
		}
	}

	return 0;
}
