/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

/* This is one of those programs that, because of a subtlety in the POSIX shell,
 * what really should be a 10 line shell script turns into a 100+ line C
 * program...  The goal is simply to start muspd if it not be already running
 * and exec lnudg to communicate with it; however, there is a race condition if
 * that be done naively and it really did occur so I wrote this program. */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <wait.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <perrf.h>
#include <time.h>
#include <pwd.h>

#ifndef MUSPD
#define MUSPD "muspd"
#endif

const char * const muspd = MUSPD;
char *muspd_unix;
char *lockfile;
char *lockdir;
int lockfile_fd;

int muspd_unix_init()
{
	const char muspd_unix_name[] = "muspd.unix";
	struct passwd *pw = getpwuid(getuid());
	muspd_unix = malloc(strlen(pw->pw_dir) + 1
		+ strlen(muspd_unix_name) + 1);
	if (muspd_unix == NULL)
		return 1;
	stpcpy(stpcpy(stpcpy(muspd_unix, pw->pw_dir), "/"), muspd_unix_name);
	
	return 0;
}

int lock_init()
{
	char id[(CHAR_BIT * sizeof(uid_t)) / 3 + 2 + 1 + 1];
	const size_t id_len = snprintf(id, sizeof id, "%llu",
			(long long unsigned)getuid());
	const char prefix[] = "/tmp/muspll.";
	const size_t lockdir_len = strlen(prefix) + id_len + 1;
	lockdir = malloc(lockdir_len + 1);
	if (lockdir == NULL)
		return 1;
	stpcpy(stpcpy(stpcpy(lockdir, prefix), id), "/");
	const char lock[] = "lock";
	const size_t lockfile_len = lockdir_len + strlen(lock) + 1;
	lockfile = malloc(lockfile_len + 1);
	if (lockfile == NULL)
		return 1;
	stpcpy(stpcpy(lockfile, lockdir), lock);

	return 0;
}

int lock()
{
	if (mkdir(lockdir, 0700)) {
		if (errno != EEXIST) {
			perrf("failed to create lock directory %s",
					lockdir);
			return 1;
		}
	}
	const int lockdir_fd = open(lockdir, O_DIRECTORY | O_RDONLY);
	if (lockdir_fd == -1) {
		perrf_errno("failed to open lock directory %s",
				lockdir);
		return 1;
	}
	struct stat st;
	if (fstat(lockdir_fd, &st))
		return 1;
	if (st.st_uid != getuid()) {
		perrf("lock directory %s already exists "
				"and is not owned by thee",
				lockdir);
		return 1;
	}
	if (st.st_mode & 077) {
		perrf("lock directory %s has too permissive mode", lockfile);
		return 1;
	}
	lockfile_fd = openat(lockdir_fd, lockfile, O_CREAT | O_RDWR,
			0600);
	if (lockfile_fd == -1) {
		perrf_errno("failed to open lock file %s", lockfile);
		return 1;
	}
	if (close(lockdir_fd))
		return 1;
	struct flock fl = {
		.l_type = F_WRLCK,
		.l_whence = SEEK_SET,
		.l_start = 0,
		.l_len = 0
	};
	if (fcntl(lockfile_fd, F_SETLKW, &fl)) {
		perrf_errno("failed to acquire lock on file %s", lockfile);
		return 1;
	}

	return 0;
}

int unlock()
{
	if (close(lockfile_fd)) {
		perrf_errno("failed to close lock file %s", lockfile);
		return 1;
	}

	return 0;
}

int muspd_start()
{
	if (access(muspd_unix, F_OK) != 0) {
		pid_t pid = fork();
		int res;
		switch (pid) {
			case 0:
				execlp(muspd, muspd,
						muspd_unix,
						(char *)NULL);
				perrf_errno("failed to execute %s", muspd);
				exit(1);
			case -1:
				perrf_errno("failed to fork");
				return 1;
			default:
				if (waitpid(pid, &res, 0) != pid) {
					perrf_errno("failed to wait for muspd");
					return 1;
				}
				if (WIFEXITED(res) && WEXITSTATUS(res) == 0) {
					return 0;
				} else {
					perrf("muspd returned non-zero");
					return 1;
				}
				break;
		}
	}

	return 0;
}

int sig_init()
{
	struct sigaction act = { 0 };
	act.sa_handler = SIG_IGN;
	if (sigemptyset(&act.sa_mask))
		return 1;
	if (sigaction(SIGINT, &act, NULL))
		return 1;
	if (sigaction(SIGTERM, &act, NULL))
		return 1;
	if (sigaction(SIGQUIT, &act, NULL))
		return 1;

	return 0;
}

int main(int argc, char *argv[])
{
	perrf_head = argv[0];

	int stay_alive = 0;
	int opt;
	while ((opt = getopt(argc, argv, "a")) != -1) {
		switch (opt) {
			case 'a':
				stay_alive = 1;
				break;
			default:
				perrf("usage: muspll [-a]");
				break;
		}
	}

	if (sig_init())
		return 1;
	if (muspd_unix_init())
		return 1;
	if (lock_init())
		return 1;
	if (lock())
		return 1;
	if (muspd_start()) {
		unlock();
		return 1;
	}
	if (unlock())
		return 1;

	int lnudg_argc = 0;
	char *lnudg_argv[4];
	lnudg_argv[lnudg_argc++] = "lnudg";
	if (stay_alive)
		lnudg_argv[lnudg_argc++] = "-a";
	lnudg_argv[lnudg_argc++] = (char *)muspd_unix;
	lnudg_argv[lnudg_argc++] = NULL;

	return execvp("lnudg", lnudg_argv);
}
