/*
 * Copyright (C) 2019-2024 B.R.S. Roso.
 * 
 * Musp is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Musp.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "perrf.h"

/* This is a small library used to print error messages. The program using it
 * sets, if desired, the global variable perrf_head to, say, the program's
 * name; then, subsequent calls to perrf will cause it to print perrf_head
 * followed by ": "; then, the arguments are used to construct the rest of the
 * string just as if printf had been called. The function perrf_errno does the
 * same as perrf initially, then it calls perror with perrf_head as argument so
 * that the error message relating to the current value of errno be printed. In
 * either function, if perrf_head be NULL (initial value) then the header is
 * simply not printed; ie, the message won't be started by ": ".
 *
 * This could be done more easily, but an effort is made to try and print an
 * entire line with a single call to write. Thus the output should look nicer
 * in the event many programs be writing to the same terminal at the same time.
 */

const char *perrf_head;

/* Write perrf_head followed by ": " onto the buffer str. Return pointer to the
 * terminating null character. */
static char *perrf_head_make(char *str)
{
	char *p = str;
	if (perrf_head) {
		const size_t perrf_head_len = strlen(perrf_head);
		memcpy(p, perrf_head, perrf_head_len);
		p += perrf_head_len;
		*p++ = ':';
		*p++ = ' ';
	}
	*p = '\0';

	return p;
}

/* Size needed to accomodate the header excluding terminating null. */
static size_t perrf_head_made_len()
{
	if (perrf_head)
		return strlen(perrf_head) + 2;
	else
		return 0;
}

void vperrf(const char *fmt, va_list ap)
{
	if (fmt == NULL)
		return;
	
	va_list ap2;
	va_copy(ap2, ap);

	/* Need a block here because there is a goto preceding the variable
	 * length array. */
	{
		const int msg_len = vsnprintf(NULL, 0, fmt, ap);
		if (msg_len < 0)
			goto end;
		const size_t len = perrf_head_made_len() + msg_len + 1;
		char str[len + 1];
		char *p = perrf_head_make(str);
		vsnprintf(p, msg_len + 1, fmt, ap2);
		p += msg_len;
		*p++ = '\n';
		ssize_t n;
		size_t i = 0;
		while ((n = write(2, str + i, len - i))) {
			if (n < 0)
				goto end;
			i += n;
		}
	}

end:
	va_end(ap2);
}

void perrf(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vperrf(fmt, ap);
	va_end(ap);
}

void vperrf_errno(const char *fmt, va_list ap)
{
	/* Not sure what size to use here;
	 * glibc uses 1024 in perror,
	 * so I'm doing the same... */
	char buf[1024];
	const int res = strerror_r(errno, buf, sizeof buf);
	vperrf(fmt, ap);
	if (res == 0)
		perrf("%s", buf);
}

void perrf_errno(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vperrf_errno(fmt, ap);
	va_end(ap);
}
